
//	Copyright (c) 1982 Dyadic Systems Limited	

//	Native file functions	

#include	"apl.h"
#include	"support.h"
#include	"io_maps.h"
#include	"quadav.h"
#include	<errno.h>
#include	<sys/types.h>
#include	<stdio.h>
#include	<signal.h>
#include	<sys/stat.h>
#ifdef	BSD				// Berkeley moved it!		
#include	<sys/time.h>
#else
#include	<time.h>
#endif	// BSD 
#if	defined(WINDOWS3)
#	include <share.h>
#endif

unsigned char taskid[] = "nfiles";

#ifndef	_NFILE
#define	_NFILE	20
#endif	// _NFILE 

// #undef	EBADF
// #define	EBADF	9		// Bad File Number		
// #define	EPERM	13		// Permission Denied		

#if	!defined(EBADF)
#	define	EBADF	9
#endif

#if	!defined(EPERM)
#	if	defined(EACCES)
#		define	EPERM	EACCES
#	else
#		define	EPERM	13
#	endif
#endif


#ifndef	SEEK_SET			// haven't got ANSI fseek defs	
#define	SEEK_SET	0
#define	SEEK_CUR	1
#define	SEEK_END	2
#endif	// SEEK_SET 

#define	REG	register
#define	UNS	unsigned
#define	UCHAR	unsigned char

#define	CHARS	0
#define	WORDS	1
#define	LINES	2

extern UCHAR	*malloc();
extern UCHAR	*realloc();

//	translatey stuff	
extern jmp_buf		xbufover;
extern void		i_filbuf(void);
extern void		o_filbuf(void);
extern int		test_file;
extern int		trans_no_error_messages_please;
extern struct Fd_map	in_fd[_NFILE];
extern struct Out_fd	*o_out[_NFILE];
extern void		(*x_error)(int);
extern void		(*fname)(unsigned char *, unsigned char *, unsigned);
extern void		i_fname(unsigned char *, unsigned char *, unsigned);
extern void		o_fname(unsigned char *, unsigned char *, unsigned);
struct Fastmap
	{
	unsigned char result[256];
	unsigned char absorb[256];
	};

struct Fastout		// help speed up translated output 
	{
	unsigned char index[256];
	unsigned char codes[256];
	} *f_out[_NFILE], fastout;

static ARRAY	intrslt;	// quick array for integer result	
static ARRAY	*retval();

static jmp_buf	err_env;	// handle i/o failure with set/longjmp	
static ARRAY	*Rslt;		// handle i/o failure with global ptr	
static UNS	 Rlen;		// nelts currently installed in Rslt	
static UNS	 Rtot;		// for out failure			

static FILE	*mystr[_NFILE];	// streams, indexed by fd		
static UCHAR	myfd[_NFILE];	// fd flags with the following values ..
#define	MYOP	0x01		// I've opened this fd			
#define	MYRD	0x02		// For reading				
#define	MYWR	0x04		// For writing				

static ARRAY	*xclose(),
		*xlate(),
		*xlseek(),
		*xopen(),
		*xread(),
		*xwrite();

static struct FUNCTION
		{
		unsigned char	*f_name;
		UNS	f_syntax;
		ARRAY	*(*f_pointer)();
		}functions[] =
			{
		"close",	SHYR|MONA, xclose,
		"lseek",	RSLT|OPTL, xlseek,
		"open" ,	RSLT|OPTL, xopen,
		"read" ,	RSLT|OPTL, xread,
		"write",	RSLT|OPTL, xwrite,
		"xlate",	SHYR|MONA, xlate
			};

#define	NFUNCTIONS	(sizeof(functions)/sizeof(struct FUNCTION))

#if	defined(DOS) || defined(WINDOWS3)

typedef enum
	{
	Compatibility,
	Deny_read_write,
	Deny_write,
	Deny_read,
	Deny_none
	} Sharing_mode_type;
#endif

#ifdef	DOS
#undef	data		// from support.h (clashes with pragma) 
#include <implement.cf>
/*
 * Hook for sharing across networks
 *
 * Metaware provides support for this via a global variable
 * Sharing_mode
 */
// Use data pragma to share with Pascal system.pf package: 

pragma data(common,"?" _Private_prefix "system");
	Sharing_mode_type Sharing_mode;
pragma data;
#endif	// DOS 

/*	c l e a n u p

	tidy up when killed by apl.

	cleanly close all output streams with translations in effect.
*/

static void	cleanup()
{
	REG int	fd;

	for(fd = 0; fd < _NFILE; ++fd)
		if(MYOP & myfd[fd])		// if I opened it 
			xclose(0, retval(fd));

#ifdef	DOS
	toapl(0);
	fromapl();
#endif	// DOS 

	exit(0);
}

#if	defined(DOS)
	main()
#elif	defined(WATCOM)
	win_main(int argc, unsigned char *argv[])
#else
	main(argc, argv)
	int	argc;
	unsigned char	*argv[];
#endif	// DOS 
{
	REG ARRAY	*rslt, *larg, *rarg;
	REG		fc;
	ARRAY		*getarray(), *freearray();

	startup(NFUNCTIONS);	// signal number of functions to apl 
	trans_no_error_messages_please = 1;	// else trans scribbles 

	for(fc = 0; fc < NFUNCTIONS; ++fc)
		define(functions[fc].f_name, functions[fc].f_syntax, fc);

	larg = rarg = rslt = Rslt = 0;

	setjmp(errbuff);			// come here after error 

	for(;;) 
		{
		larg = freearray(larg);		// free up heap space	
		rarg = freearray(rarg);
		if (rslt != Rslt) freearray(Rslt);	// dont do it twice
		if (rslt != &intrslt) freearray(rslt);
		rslt = Rslt = 0;

		fc = fromapl();			// get desired function	

		if(fc == -1)
			cleanup();
		else
			switch(functions[fc].f_syntax & DYAD)
				{
			case DYAD:		// dyadic & ambivalent 
				larg = getarray();
			case MONA:		// monadic 
				rarg = getarray();
				break;
			default:		// niladic 
				break;
				}
		
		rslt = (*functions[fc].f_pointer)(larg, rarg);

		putarray(rslt);			// return result	

		}
}

/*	r e t v a l

	return integer scalar array based on result of system call ...
	error if val < 0.
*/

static ARRAY	*retval(n)
int	n;
{
	if(n < 0)
		filerror();
	intrslt.length	 = 3;
	intrslt.type	 = SIMPLE;
	intrslt.rank	 = 0;
	intrslt.eltype	 = DINT;
	intrslt.shape[0] = (UNS)n;

	return &intrslt;
}

/*	g e t f i l e n a m e

	return pointer to file name string - or error
*/

static UCHAR	*getfilename(array)
ARRAY	*array;
{
	static UCHAR	filename[256];
	UCHAR 		avtounix();
	REG UCHAR	*to, *from;
	REG UNS		len;

	if(CHAR != array->eltype)
		error(DOMAIN);

	if(array->rank > 1)
		error(RANK);

	len = arraybound(array);

	if(len >= sizeof(filename))
		error(LENGTH);

	to   = filename;
	from = (UCHAR *)&array->shape[array->rank];

	while(len--)
		*to++ = avtounix(*from++);

	*to = '\0';

	return filename;
}

/*	g e t p n t r v e c

	check that 'array' is a PNTR vector with between 'min' & 'max' elts.
	return ptr to first elt.
*/

static ARRAY	**getpntrvec(array, min, max)
REG ARRAY	*array;
UNS		min;
UNS		max;
{
	if(array->eltype != PNTR)
		error(DOMAIN);

	if(array->rank != 1)
		error(RANK);

	if(array->shape[0] < min
	|| array->shape[0] > max
	  )
		error(LENGTH);

	return (ARRAY **)&array->shape[1];
}

/*	g e t n u m v e c

	check that 'array' is a numeric scalar or vector with between 'min'
	and 'max' elements.
	fill 'vec' with those thingies - unused set to 0.
*/

static void	getnumvec(array, min, max, vec)
ARRAY	*array;
UNS	min, max;
REG int	*vec;
{
	int		(*initdint())();
	REG int		(*makedint)();
	REG UNS		bnd;
	REG UCHAR	*cptr;

	if(array->eltype == DOUB)
		error(DOMAIN);

	bnd = arraybound(array);

	if(bnd < min || bnd > max)
		error(LENGTH);

	if(array->rank > 1)
		error(RANK);

	memset((UCHAR *)vec, 0, (sizeof(int) * max));	// zap to 0 
		
	for(
		cptr   = (UCHAR *)arraystart(array),
		makedint = initdint(array);

		bnd--;
	   )
		{
		vec[bnd] = (*makedint)(cptr, bnd);
		}
}

/*	g e t i n t

	return the value of a simple integral scalar or singleton vector
	- or domain error
*/

static int	getint(array)
ARRAY	*array;
{
	int	val;

	getnumvec(array, 1, 1, &val);

	return val;
}

/*	c h k r n g

	check that 'thing' is in the range min .. max
*/

static int	chkrng(thing, min, max)
REG int	thing;
int	min;
int	max;
{
	if(thing < min || thing > max)
		error(DOMAIN);
	return thing;
}

/*	c h k f d

	check that the user-supplied file-descriptor is a legal value - ie
	not ARGSPIPE, RSLTPIPE, ...?

	also check that i've opened it, in the appropriate mode
	(eg 0, MYRD, MYWR)
*/

static int	chkfd(fd, mode)
REG int	fd;
int	mode;
{
	switch(chkrng(fd, 0, _NFILE))
		{
#if	defined(DOS) || defined(WINDOWS3)
#else
	case ARGSPIPE:			// APL->AP pipe 
	case RSLTPIPE:	error(DOMAIN);	// AP->APL pipe 
#endif	// defined(DOS) || defined(WINDOWS3) 

	default	     :			// Check I've opened it	
			if(0 == (MYOP & myfd[fd]))
				errno = EBADF, filerror();
			if(0 != mode	// mode to check 
			     && 0 == (mode & myfd[fd]) // not there 
			       )
				errno = EPERM, filerror();
			break;
		}
	return fd;
}

/*	c l o s e

	z <- close fildes
*/

static ARRAY	*xclose(larg, rarg)
ARRAY	*larg, *rarg;
{
	int	fd, status;

	fd = chkfd(getint(rarg), 0);		// check it's open	

	status = fclose(mystr[fd]);

	if(status == 0)				// succeeded.		
		{
		void	relitab(), relotab();

		relitab(fd);
		relotab(fd);

		myfd[fd] = 0;			// forget about it	
		}

	return retval(status);
}

/*	s e e k r e a d

	called with xlseek's args ... read the stuff, as a slow lseek.

	returns flag saying whether it read all it should've.

	NB relies on the fact that xread's rarg is the same as the 2 take
	of xlseek's rarg, and the largs are the same too.
	therefore, the same arrays can be used and are automatically freed
	even on error.
*/

static int	seekread(larg, rarg, nitems)
ARRAY	*larg, *rarg;
int	nitems;
{
	ARRAY	*rslt;
	int	ok;

	rarg->shape[0] = 2;	// happy coincidence ... args are similar 

	rslt = xread(larg, rarg);		// read stuff		

	ok = (nitems == rslt->shape[0]);	// maybe hit eof?	

	freearray(rslt);			// & discard		

	return ok;				// say whether it worked
}

/*	l s e e k

	position <- type lseek fildes offset from

	like library function but extended to words and lines
*/

static long	snt(fildes, offset, from)	// seek 'n' tell 
int	fildes;
long	offset;
int	from;
{
	long	pos;
	FILE	*stream;

	stream = mystr[fildes];

	pos = fseek(stream, offset, from);

	if(pos == 0)	// worked ok 
		pos = ftell(stream);

	return pos;
}

static ARRAY	*xlseek(larg, rarg)
ARRAY	*larg, *rarg;
{
	int	type;
	int	rvec[3];
	int	fd;
	long	offset;
	int	from;
	int	xlate;

	type = chkrng(larg ? getint(larg) : CHARS, CHARS, LINES);

	getnumvec(rarg, 3, 3, rvec);

	fd     = chkfd(rvec[0], 0);		// check it's open	

	offset = rvec[1];

	switch(chkrng(rvec[2], 0, 2))
		{
	case 0:	from = SEEK_SET;	break;
	case 1:	from = SEEK_CUR;	break;
	case 2:	from = SEEK_END;	break;
		}

	xlate = (isx(fd, 0) || isx(fd, 1));	// any translate ? 

	if(xlate || type != CHARS)		// not raw CHARS	
		if((offset < 0		    ||	// can't seek back	
	    	    offset > 0 && from == 2	// or forward from end	
		  ))
			error(NONCE);
		else if(offset)			// legal non-raw-CHAR	
			{			// have to read along	
			long	wasat;
			struct Fd_map	save;	// might read across mode 
						// change, then seek back!

			wasat = snt(fd, 0L, 1); // save current posn	
			save = in_fd[fd];	  // save state		

			if(-1 == wasat)		// can't seek on fd	
				filerror();

			snt(fd, 0L, from);	// initial posn		

			if(seekread(larg, rarg, offset))// success	
				{
				from   = SEEK_CUR;
				offset = 0;	// return new posn	
				}
			else				// failure	
				{
				in_fd[fd] = save; // restore state	
				from   = SEEK_SET;	// reset to ...	
				offset = wasat;		// original posn
				}
			}

	// ... easy cases drop thru to simple snt ...		

	return retval(snt(fd, offset, from));	// report position or do it 
}
	
/*	o p e n

	fildes <- {mode {create}} open name
	or
	fildes <- {textmode {share}} open name

	Default mode is read-only.
	Default create flag is don't.

	if create flag is 0, file must already exist.
	if		  1, it must not exist.

	Disadvantage of junking creat is that you can't specify the access.
*/

static int	fexist(name)
unsigned char	*name;
{
	FILE	*stream;

	stream = fopen(name, "r");	// fails if file don't exist 

	if(NULL == stream)
		return 0;
	else
		{
		fclose(stream);

		return 1;
		}
}

static void	fcreate(name)
unsigned char	*name;
{
	FILE	*stream;

	stream = fopen(name, "a");

	if(NULL == stream)
		filerror();
	else
		fclose(stream);
}

#if	defined(DOS) || defined(WINDOWS3)
static Sharing_mode_type getshare(ARRAY *arg)
{
	Sharing_mode_type sht;
	int vec[1];

	getnumvec(arg, 1, 1, vec);
	switch(sht = (Sharing_mode_type)vec[0])
		{
	case Compatibility: case Deny_read_write:
	case Deny_write: case Deny_read: case Deny_none:
		break;
	default: error(DOMAIN);
		}
	return sht;
}
#endif	// DOS 

static unsigned char *gettype(arg)
ARRAY * arg;
{
	unsigned char *cp;
	unsigned awr = 0;	// flags for a w r   
	unsigned pbtu = 0;	// flags for + b t u 
	unsigned bnd;
	static unsigned char buf[6];
	unsigned char *bp = buf;

	if(arg->eltype != CHAR) error(DOMAIN);
	if(arg->rank > 1) error(RANK);
	bnd = arraybound(arg);
	cp = (unsigned char *)arraystart(arg);
	while(bnd--) switch(*cp++)
		{
	case RL:
		if(awr) error(DOMAIN);// only set one bit and not repeat it 
		awr = 1;
		*bp++ = 'r';
		break;
	case WL:
		if(awr) error(DOMAIN);// only set one bit and not repeat it 
		awr = 2;
		*bp++ = 'w';
		break;
	case AL:
		if(awr) error(DOMAIN);// only set one bit and not repeat it 
		awr = 4;
		*bp++ = 'a';
		break;
	case PLUS:
		if(pbtu & 8) error(DOMAIN); // no repeats 
		pbtu |= 8;
		*bp++ = '+';
		break;
	case TL:
		if(pbtu & 4) error(DOMAIN); // no repeats 
		pbtu |= 4;
		*bp++ = 't';
		break;
	case BL:
		if(pbtu & 2) error(DOMAIN); // no repeats 
		pbtu |= 2;
		*bp++ = 'b';
		break;
	case UL:
		if(pbtu & 1) error(DOMAIN); // no repeats 
		pbtu |= 1;
		*bp++ = 'u';
		break;
	default: error(DOMAIN);
		}
	if(awr == 0) error(DOMAIN);
	*bp = 0;
	return buf;
}


static ARRAY	*xopen(larg, rarg)
ARRAY	*larg, *rarg;
{
	unsigned char	*type;
	int		fd;
	FILE		*stream;
	unsigned char	*fn;
	int		mymo = 0;


#if	defined(DOS) || defined(WINDOWS3)
	Sharing_mode_type	sht = Compatibility;
#endif	// DOS 

	switch (larg ? larg->eltype : INTG)
		{
	case PNTR:
		if(larg->rank != 1) error(RANK);
		if(larg->shape[0] != 2) error(LENGTH);
		type = gettype((ARRAY *)larg->shape[1]);
		sht = getshare((ARRAY *)larg->shape[2]);
		fn = (unsigned char *)getfilename(rarg);
		break;
	case CHAR:
		type = gettype(larg);
		fn = (unsigned char *)getfilename(rarg);
		break;
	default:
		{
		int	vec[2];
		getnumvec(larg ? larg : retval(0), 0, 2, vec);


		switch(vec[0])
			{
		case 0 :
			mymo = MYRD;
#ifdef	ANSI
			type = "rb";
#else
			type = "r";	// eg 6150 ... doesn't recognise "b" 
#endif	// ANSI 
			break;
		case 1 :		// all write-only fopens trunc file 
			mymo = MYWR;
		case 2 :
			if(0 == mymo) mymo = MYWR | MYRD;
#ifdef	ANSI
			type = "rb+";
#else
			type = "r+";
#endif	// ANSI 
			break;

		default:
			error(DOMAIN);
			break;
			}

		fn = (unsigned char *)getfilename(rarg);

		switch(vec[1])
			{
		case 0:			// ensure file already exists	
			if(!fexist(fn))
				filerror();

			break;

		case 1:			// ensure file doesn't exist	
			if(fexist(fn))
				error(DOMAIN);
			else
				fcreate(fn);	// errors if can't	

			break;
			}
		}
		}

#if	defined(DOS)
	Sharing_mode = sht;
	stream = fopen(fn, type);
	Sharing_mode = Compatibility;
#elif	defined(WINDOWS3)
		{
		static int share[] =
			{
			SH_COMPAT
		,	SH_DENYRW
		,	SH_DENYWR
		,	SH_DENYRD
		,	SH_DENYNO
			};
		stream = _fsopen(fn, type, share[sht]);
		}
#else
	stream = fopen(fn, type);
#endif	// DOS 

	if(stream != NULL)
		{
		fd = fileno(stream);
		mystr[fd] = stream;

		if(0 == mymo)
			{
			if     (strchr(type, '+')) mymo = MYRD | MYWR;
			else if(strchr(type, 'r')) mymo = MYRD;
			else if(strchr(type, 'w')) mymo = MYWR;
			else if(strchr(type, 'a')) mymo = MYWR;
			}

		myfd[fd] = MYOP | mymo;	// log fd in private table 
		}
	else
		fd = -1;

	return retval(fd);
}

//**********************************************************************
//		I N P U T   S T U F F					
//**********************************************************************

static int	(*inch)();
static int	in_fld;
static FILE	*in_sd;

static UCHAR	*b_fst = 0;
static UCHAR	*b_ptr = 0;
static UCHAR	*b_end = 0;

#define	STRINCR	128		// word/lin buf increment 

static void	build(which, ch)
int		which;
UCHAR	ch;
{
	switch(which)
		{
	case 0:					// zap 
		if(b_fst)
			{
			free(b_fst);

			b_fst = b_ptr = b_end = 0;
			}
		break;

	case 1:					// add			
		if(b_fst)			// already a buf	
			{
			if(b_ptr == b_end)	// but it's full	
				{
				UNS len;

				len = STRINCR + (b_end - b_fst);

				b_fst = realloc(b_fst, len);
				b_ptr = b_fst + len - STRINCR;
				b_end = b_fst + len;
				}
			}
		else				// first buf		
			{
			b_fst = malloc(STRINCR);
			b_ptr = b_fst;
			b_end = b_fst + STRINCR;
			}

		*b_ptr++ = ch;
		}
}

/*	b u i l d a r r a y

	form an array from the stuff in the build buffer.
*/

static ARRAY	*buildarray()
{
	ARRAY	*rslt;
	UNS	len;

	len = b_ptr - b_fst;		// yuk 

	rslt = mkarray(CHAR, 1, &len);

	memcpy(arraystart(rslt), b_fst, len);

	return rslt;
}

/*	i _ f a i l

	input's gone wrong somehow.
*/

#define	NWDS(nchars)	((nchars + sizeof(int) - 1) / sizeof(int))

static void	i_fail()
{
	int err = errno;
	
	clearerr(in_sd);		// in case we can continue to  read 
	if(Rslt && err == 0)		// got at least some way	
		{
		UNS	*shape;

		shape = Rslt->shape;

		switch(Rslt->eltype)
			{
		case CHAR:
			Rslt->length -= NWDS(shape[0]) - NWDS(Rlen);
			break;

		case PNTR:
			if(b_fst)		// partial item built 
				{
				*(ARRAY **)&shape[1+Rlen] = buildarray();

				++Rlen;
				}

			if(!Rlen)		// fill with prototype 
				*(ARRAY **)&shape[1] = mkarray(CHAR, 0, 0);

			Rslt->length -= shape[0] - (Rlen ? Rlen : 1);

			build(0, 0);		// free partial item space 

			break;
			}

		shape[0] = Rlen;		// correct length	

		longjmp(err_env, 1);
		}
	else				// fail				
		{
		freearray(Rslt);

		filerror();
		}
}

/*	s p e c m a p

	map a special (ie negative) character to a plausible quadav position.
*/

static int	specmap(special)
int	special;
{
	switch(special)
		{
	case  ER: return NEWLINE;		// Enter 
	case  TB: return SP;			// Horizontal Tab 
	case  RC: return SP;			// Right Character 
	case  LC: return BS;			// Left Character 
	case  DP: return SP;			// Destructive Space 
	case  DB: return BS;			// Destructive Backspace 
	case  NB: return BS;			// Non Destructive Backspace 
	case  NS: return SP;			// Non Destructive Space 

	default : return special * (special >= 0);
		}
}

static int	isdelim(ch, type)
REG int	ch;
int	type;
{
	switch(type)
		{
	default:			// there are not any other cases
	case CHARS:	return 0;
	case WORDS:	return (ch == SP || ch == NEWLINE);
	case LINES:	return (ch == NEWLINE);
		}
}

/*	i n x c h a r

	read the next char ("keystroke") from fd using translation.
*/

static int	inxchar()
{
	REG int	key;

	key = inkey(in_fld, USER);

	if(key == 0)			// end of file	
		i_fail();
	return (key < 0) ? specmap(key) : key;
}

static int infchar()	 	// defensive : should never be called 
{
	return inxchar();
}

/*	i n c h a r

	read the next char from fd without using translation.
*/

static int	inchar()
{
	REG int	ch;

	ch = fgetc(in_sd);
	if(ch == EOF)		// end of file	
		i_fail();
	return ch;
}

/*	i n w o r d

	read a word from fd - translated or not.
*/

static void	inword()
{
	REG UCHAR	ch;

	do				// skip leading delims	
		{
		ch = (*inch)();
		}
	while(isdelim(ch, WORDS));

	// ch is first character of word ... 

	build(0, 0);

	while(!isdelim(ch, WORDS))
		{
		build(1, ch);

		ch = (*inch)();
		}
}

/*	i n l i n e

	read a line from fd - translated or not.
*/

static void	inline()
{
	REG int	ch;

	build(0, 0);

	for(
		ch = (*inch)();

		!isdelim(ch, LINES);

		ch = (*inch)()
	   )
		build(1, ch);
}

static int	disk_read(fd, wait)	// speed up inkey for disk reads 
int	fd;
int	wait;
{
	REG int	ch;

	ch = fgetc(in_sd);

	return (ch == EOF) ? -1 : 0xff & ch;
}

/*	p r e p i n

	prepare for input on file descriptor 'fd'.
*/

static void	prepin(fd)
int	fd;
{
	if(isx(fd, 0))				// translate on 
		{
		extern int	timed_read();
		extern int	(*readfn)();
		int		save_errno = errno;

		if(isatty(fd))
			{
			readfn = timed_read;
			inch = inxchar;
			}
		else
			{
			readfn = disk_read;
			inch =  is_fast(fd) ? infchar : inxchar;
			}
		errno = save_errno;
		}
	else
		inch = inchar;

	in_fld = fd;
	in_sd  = mystr[fd];
}

/*	d o n e i n

	just finished input on file descriptor 'in_fld'.
*/

static void	donein()
{
}

/*	r e a d c h a r s
*/

static void	readchars(nchars)
int	nchars;
{
	UCHAR	*ptr;

	Rslt = mkarray(CHAR, 1, &nchars);

	Rlen = 0;
	ptr  = (UCHAR *)arraystart(Rslt);

	if(inch == inchar)		// speed up raw read		
		{
		Rlen = fread(ptr, 1, nchars, in_sd);

		if(Rlen != nchars)
			i_fail();
		}
	else if(inch == infchar)
		{
		Rlen = fastxread(ptr, 1, nchars, in_sd);

		if(Rlen != nchars)
			i_fail();
		}
	else				// slower translated read	
		{
		while(Rlen < nchars)
			{
			*ptr++ = inxchar();

			++Rlen;
			}
		}
}

static void	readthings(nthings, infn)
int	nthings;
void	(*infn)();
{
	REG ARRAY	**ptr;

	Rslt = mkarray(PNTR, 1, &nthings);
	ptr = (ARRAY **)arraystart(Rslt);
	if (nthings)
	{
		for(
			Rlen = 0;

			Rlen < nthings;

			++Rlen, ++ptr
		   )
			{
			(*infn)();			// build the item 

			*ptr = buildarray();		// array from built item 
			}
	}
	else
	{				//Fix bug 8555 by giving it a prototype
		*ptr = mkarray(CHAR, 1, &nthings);
	}
}

/*	r e a d w o r d s
*/

static void	readwords(nwords)
int	nwords;
{
	readthings(nwords, inword);
}

/*	r e a d l i n e s
*/

static void	readlines(nlines)
int	nlines;
{
	readthings(nlines, inline);
}

/*	r e a d

	buffer <- type read fildes, nitems
*/

static ARRAY	*xread(larg, rarg)
ARRAY	*larg, *rarg;
{
	static void	(*readfns[])() = { readchars, readwords, readlines };
	int	type;
	int	rvec[2];

	type = chkrng(larg ? getint(larg) : CHARS, CHARS, LINES);

	getnumvec(rarg, 2, 2, rvec);

	if(rvec[1] < 0)				// write, perhaps?	 
		error(DOMAIN);

	Rslt  = 0;				// so can detect prep err 
	errno = 0;

	if(!setjmp(err_env))
		{
		prepin(chkfd(rvec[0], MYRD));	// chk open for rd 

		(*readfns[type])(rvec[1]);	// set up global rslt	
		}

	donein();

	return Rslt;				// return global rslt	
}

//**********************************************************************
//		O U T P U T   S T U F F					
//**********************************************************************

static void	(*outvec)();
static void	(*flushvec)();
extern FILE	*out_sd;		// the current output stream	

/*	o _ f a i l

	output's gone wrong somehow.
*/

static void	o_fail()
{
	if(errno)
		filerror();
	else
		longjmp(err_env, 1);
}

extern void avtoasc();

void unknown(              c)
             unsigned char c;
{
	extern int (*out_fn)(); 			// from trans.c 
	extern int t_safe();				// from trans.c 
	extern unsigned char t_is_safe, t_safe_code;    // from trans.c 
	int (*save)();
	void buf_one();

	save = out_fn;
	out_fn = t_safe;
	avtoasc(c);
	out_fn = save;
	fastout.index[c] = t_is_safe;
	fastout.codes[c] = t_is_safe ? t_safe_code : c;
}

void linefeed(              c)
              unsigned char c;
{
	unknown(NEWLINE);	// side effect - zeros the current col 
	avtoasc(c);		// put out the LINEFEED 
}


static void	outtrans(ptr, bnd, term)
REG UCHAR	*ptr;
UNS		bnd;
UCHAR		term;
{
	REG UNS	cnt;
	typedef void (*VOIDFN)();
	static VOIDFN	vfns[] =
		{
		(VOIDFN)avtoasc,
		(VOIDFN)buf_one,
		(VOIDFN)linefeed
		};

	for(cnt = bnd; cnt--; ++ptr)
		(*vfns[fastout.index[*ptr]])(fastout.codes[*ptr]);

	Rtot += bnd;

	if(term)
		{
		avtoasc(term);

		++Rtot;
		}
}

static int	right(buf, nbytes)
unsigned char	*buf;
int	nbytes;
{
	REG int	wrote;

	wrote = fwrite(buf, 1, nbytes, out_sd);

	if(wrote < 0)
		o_fail();
	return wrote;
}

static void	outraw(ptr, bnd, term)
UCHAR	*ptr;
UNS	bnd;
UCHAR	term;
{
	REG int	wrote;


	wrote = right(ptr, bnd);		// line of text first	

	Rtot += wrote;

	if(wrote == bnd)			// line writ all ok	
		{
		if(term)			// end to word or line	
			{
			wrote = right(&term, 1);

			Rtot += wrote;

			if(wrote != 1)
				o_fail();	// end didn't work	
			}
			
		}
	else					// text didn't work	
		o_fail();
}

/*	o u t _ b u f
	output characters into a memory buffer, functions
	buf_flush(), buf_init() are associated
*/
unsigned char membuf[32*1024];
unsigned char *next_in_buffer;
int (*write_fn)();

void buf_flush()
{
	int count = next_in_buffer - membuf;

	(*write_fn)(membuf, count);
	next_in_buffer = membuf;
}

void buf_one(              c)
             unsigned char c;
{
	if(next_in_buffer == &membuf[sizeof(membuf)]) buf_flush();
	*next_in_buffer++ = c;
}

int buf_out(               cp,     count)
            unsigned char *cp; int count;
{
	/* notes.
	 * this code is only used for translated output
	 * translated output comes one APL character at a time
	 * the translate table can only cope with 255 output characters for
	 * one input character.
	 * Therefore this code never has to worry about count being larger
	 * than the size of membuf.
	 */
	if(&next_in_buffer[count] >= &membuf[sizeof(membuf)]) buf_flush();
	switch(count)
		{
	case 1:
		*next_in_buffer = *cp;
		break;
	default:
		memcpy(next_in_buffer, cp, count);
		break;
		}
	next_in_buffer += count;
	return count;
}

void buf_init()
{
	extern int (*out_fn)(); // from trans.c 

	next_in_buffer = membuf;
	write_fn = out_fn;
	out_fn = buf_out;
}

static void flushraw()
{
}

static void flushxlate()
{
	buf_flush();
}

/*	p r e p o u t

	Prepare for output via (*outvec) on file descriptor 'fd'.

	return the terminating character for the translation and type -
	0 for none.
*/

static UCHAR	prepout(fd, type)
int	fd;
int	type;
{
	int		xlate;
	static UCHAR	terms[] = { 0, SP, NEWLINE };
	FILE		*stream;

	xlate = isx(fd, 1);

	stream = mystr[fd];

	if(xlate)
		{
		init_out(stream, o_fail);
		outvec = outtrans;
		buf_init();
		flushvec = flushxlate;
		fastout = *f_out[fd];
		}
	else
		{
		out_sd = stream;
		outvec = outraw;
		flushvec = flushraw;
		}

	return terms[type];
}

/*	d o n e o u t

	Just finished output on stream out_sd
*/

static void	doneout()
{
	(*flushvec)();
	fflush(out_sd);
}

static void	outchar(array, term)
ARRAY		*array;
REG UCHAR	term;
{
	REG UNS		cols;
	REG UNS		rows;
	int		rank;
	REG UCHAR	*ptr;

	rank = array->rank;

	cols = rank ? array->shape[rank-1] : 1;

	for(
		rows = 1,
		rank -= 2;

		rank >= 0;

		--rank
	   )
		rows *= array->shape[rank];


	for(
		ptr = (UCHAR *)arraystart(array);

		rows;

		--rows, ptr += cols
	   )
		(*outvec)(ptr, cols, term);
}

static void	outpntr(array, term)
ARRAY	*array;
UCHAR	term;
{
	REG UNS		bnd;
	REG ARRAY	**ptr;

	for(
		bnd = arraybound(array),
		ptr = (ARRAY **)&array->shape[array->rank];

		bnd;

		--bnd, ++ptr
	   )
		switch((*ptr)->eltype)
			{
		case CHAR:	outchar(*ptr, term);	break;
		case PNTR:	outpntr(*ptr, term);	break;

		default  :	error(DOMAIN);		break;
			}
}

/*	w r i t e

	nchars <- type write fildes buffer
*/

static ARRAY	*xwrite(larg, rarg)
ARRAY	*larg, *rarg;
{
	int	type;
	UCHAR	term;
	ARRAY	**subvec;
	ARRAY	*dat;

	type = chkrng(larg ? getint(larg) : CHARS, CHARS, LINES);

	subvec = getpntrvec(rarg, 2, 2);

	term = prepout(				// prepare for output	 
			chkfd(			// on the valid fd in	 
				getint(		// the int singleton	 
					subvec[0]	// given here	 
				      ),
				MYWR		// chk open for write	
			     ),
			type
		      );

	dat = (ARRAY *)subvec[1];

	Rtot  = 0;			// globals to cope with failure	
	errno = 0;

	if(!setjmp(err_env))
		{
		switch(dat->eltype)
			{
		case CHAR:	outchar(dat, term);	break;
		case PNTR:	outpntr(dat, term);	break;

		default  :	error(DOMAIN);		break;
			}
		}

	doneout();

	return retval(Rtot);
}

//**********************************************************************
//			T R A N S L A T I O N S				
//**********************************************************************

/*	i s x

	is there an input (io==0) or output (io==1) translation on fd?
*/

static int	isx(fd, io)
int	fd;
int	io;
{
	switch(io)
		{
	case 0: return (0 != in_fd[fd].map);
	default:return (0 != o_out[fd]    );
		}
}

#define	BUFINCR	4096	// allocate extra 4K each time xlate read fails 

/*	r e l i t a b

	release any input table associated with fd.

	slightly tricky cos streams may share tables.

*/

static void	relitab(fd)
int	fd;
{
	REG int	i;

	if
		(
		isx(fd, 0) &&			// translate on		
		in_fd[fd].map			// table installed 	
		)
		{
		for(i = 0; i < _NFILE; ++i)	// check for synonyms
			if(in_fd[i].map == in_fd[fd].map && i != fd)
				break;

		if(i == _NFILE)			// fd is the only ref
			{
			if(in_fd[fd].map->pad[0] == 2) // fast map ? 
				free(in_fd[fd].map->mode[1]); // free it 
			free(in_fd[fd].map);	// free the space    
			}

		in_fd[fd].map = 0;
		}
}
/*	r e l o t a b

	release any output table associated with fd.

	slightly tricky cos equivalent streams probly share tables.
*/

static void	relotab(fd)
REG int	fd;
{
	REG int	i;

	if
		(
		isx(fd, 1) &&			// translate on		
		o_out[fd]->t_out		// table installed 	
		)
		{
		for(i = 0; i < _NFILE; ++i)	// check for synonyms
			if
				(
				(o_out[i] == o_out[fd] && i != fd) ||
					(
					isx(i, 1) &&
					o_out[i]->t_out == o_out[fd]->t_out
					)
				)
				break;

		if(i == _NFILE)			// fd is the only ref
			{
			free(o_out[fd]->t_out);	// free the space    

			o_out[fd]->t_out = 0;
			}

		o_out[fd] = 0;
		if(f_out[fd])
			free(f_out[fd]);
		}
}

/*	i o _ e r r o r

	cope with an error getting input or output translate tables.
*/
	
static void	io_error(n)
int	n;
{
	static void	(*tabrelfns[])() = { relitab, relotab };

	(*tabrelfns[fname == o_fname])(test_file);

	error(NFL_ERRBASE + n);
}

/*	g e t x i n p u t

	read an input translate table of the given name, associate with
	the given fd.
*/

getxinput(fd, it)
int	fd;
unsigned char	*it;
{
	relitab(fd);				// release current table 

	if(0 == strlen(it))
		return;				// cancel xlation	

	test_file = fd;				// global used by get	
	fname	  = i_fname;			// ditto		
	x_error   = io_error;			// ditto		

	if(i_get(it))
		{				// need to read file
		i_filbuf();			// try get xlate tables	
		}
}

/*	g e t x o u t p u t

	read an output translate table of the given name, associate with
	the given fd.
*/

static void	getxoutput(fd, ot)
int	fd;
unsigned char	*ot;
{
	unsigned c;

	relotab(fd);				// release current table 

	if(0 == strlen(ot))
		return;				// cancel xlation	

	test_file = fd;				// global used by get	
	fname	  = o_fname;			// ditto		
	x_error   = io_error;			// ditto		

	if(0 != o_get(ot))
		{
		o_filbuf();			// try get xlate tables	
		}

	f_out[fd] = (struct Fastout *)malloc(sizeof(struct Fastout));

	if(0 == f_out[fd])
		io_error(23);			// MALLOC_FAIL

	t_safe_init_out(fd);

	for(c = 256; c--;)
		unknown(c);
	fastout.index[LF] = 2;	// special case LINEFEED 
	*f_out[fd] = fastout;
}

/*	x l a t e

	junk <- xlate fd iotabname
	junk <- xlate fd itabname otabname
*/

static ARRAY	*xlate(larg, rarg)
ARRAY	*larg, *rarg;
{
	int	fildes;
	ARRAY	*fd, *it, *ot;
	ARRAY	**subvec;

	subvec = getpntrvec(rarg, 2, 3);

	switch(rarg->shape[0])
		{
	case 2:	 fd = subvec[0];
		 it = subvec[1];
		 ot = it;

		 break;

	case 3:	 fd = subvec[0];
		 it = subvec[1];
		 ot = subvec[2];

		 break;
		}

	fildes = chkfd(getint(fd), 0);		// check open		

	if(myfd[fildes] & MYRD)			// open for read	
		getxinput(fildes, getfilename(it));

	if(myfd[fildes] & MYWR)			// open for write	
		getxoutput(fildes, getfilename(ot));

	return retval(0);			// return junk		
}

static int is_fast(fd)
int fd;
{
	struct Imap *imap = in_fd[fd].map;

	switch(imap->pad[0])
		{
	case 0:		// not tested yet 
		break;
	case 1:		// tested and failed 
		return 0;
	case 2:		// tested and good 
		return 1;
		}
	imap->pad[0] = 1;	// set to bad 
	if(imap->n_modes != 1)
		return 0;	// only support single mode 

	{
	struct Mmap *m_map = imap->mode[0];
	unsigned char *sp = m_map->strs;
	unsigned char *dp = m_map->direct;
	short *cp = m_map->char_or_index;
	short *rp = m_map->result;
	short rows = m_map->rows;
	short cols = m_map->cols;
	short i = rows;
	unsigned char marker[256];
	struct Fastmap fm, *mfm;

	memset(marker, 0, sizeof(marker));
	for(; i--; sp += cols) 
		{
		if(marker[*sp])
			return 0; // ambiguous start of string char 
		marker[*sp] = 1;
		}
	memset(&fm, 0, sizeof(struct Fastmap));
	sp -= rows*cols;
	memset(&fm.absorb[0], 1, 256);
	for(i = 256; i--;)
		{
		if(dp[i >> 3] & (0x80 >> (i & 7)))
			{
			fm.result[i] = specmap(cp[i]);
			}
		else
			{
			fm.absorb[i] = strlen(sp + cp[i]*cols);
			fm.result[i] = specmap(rp[cp[i]]);
			}
		}
	if(m_map->doistrip)
		{
		memcpy(&fm.result[128], fm.result, 128 * sizeof(*fm.result));
		memcpy(&fm.absorb[128], fm.absorb, 128);
		}
	mfm = (struct Fastmap *) malloc(sizeof(struct Fastmap));
	if(0 == mfm)
		return 0;		// cant get space for map 
	memcpy(mfm, &fm, sizeof(struct Fastmap));
	imap->mode[1] = (struct Mmap *)mfm;
	imap->pad[0] = 2;	// mark it as good 
	return 1;
	}
}

int fastxread(               buf,     width,     nitems,       stream)
              unsigned char *buf; int width; int nitems; FILE* stream;
{
	int count = width*nitems;
	unsigned char * to = buf;
	unsigned char * from;
	unsigned char * lim = buf + count;
	struct Fastmap *fm = (struct Fastmap *)in_fd[in_fld].map->mode[1];
	unsigned char res;
	int got;

	for(;to < lim; count = lim - to)
		{
		got = fread(from = to, 1, count, stream);
		lim = to + got;
		for(; from < lim ;)
			{
			res = fm->result[*from];
			from += fm->absorb[*from];
			*to = res;
			to += 0!= res;
			}

		if(count != got) break; // give up following a short read 

		if(from > lim)	// absorbed several across buffer boundary 
			{	// avoid using fseek() in case its a pipe 
			unsigned char junk[256];
			unsigned count = from - lim;
			unsigned get;

			while(got && count)
				{
				get = count;
				if(get > sizeof(junk)) get = sizeof(junk);
				got = fread(junk, 1, get, stream);
				count -= got;
				}
			}
		}
	return to - buf;
}

#ifdef XAPL
xerror(msg)	// to support code starting X window system parts 
unsigned char *msg;
{
	unsigned char buff[64];
	sprintf(buff, "nfiles : %s\n", msg);
	write(2, buff, strlen(buff));
	exit(0);
}

void setXwindow(win)
{
}

#endif // XAPL 

//	Copyright (c) 1982 Dyadic Systems Limited	
